/*
 * db.c - fsf
 *
 * last modified: 2000-02-06
 *
 * this is based on some source I've read and is free
 */ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <sys/time.h>
#include <unistd.h>
#include <oratypes.h>
#include <ociapr.h>
#include <ocidem.h>
#include "fsql.h"

Lda_Def lda;
Cda_Def cda;
ub4     hda[HDA_SIZE/(sizeof(ub4))];
text bind_values[MAX_BINDS][MAX_ITEM_BUFFER_SIZE];
text sql_statement[2048];

DescribeStruct desc[MAX_SELECT_LIST_SIZE];
DefineStruct def[MAX_SELECT_LIST_SIZE];

extern char app_name[FILENAME_MAX];
extern sword stdin_is_tty;
extern sword stdout_is_tty;
extern sword display_mode;
extern sword display_col_names;
extern sword display_timings;
extern ConfigStruct cfg;

// cached rows
char ***row;
long nrows;
sword max_col_len[MAX_SELECT_LIST_SIZE];

sword strip_string (char *s)
{
	sword len;
	sword i;

	if (! s) {
		return -1;
	}

	len = strlen (s);

	for (i = (len - 1); i >= 0; --i) {
		if (isspace (s[i])) {
			s[i] = '\0';
		}
		else {
			break;
		}
	}

	return (strlen (s));
}

sword logon (void)
{
	text *username;
	text *password;
	int i;

	for (i = 0; i < 3; ++i) {
		if ((! stdin_is_tty) && (i > 0)) {
			break;
		}
		if ((i == 0) && (cfg.username[0] || cfg.password[0])) {
			username = malloc (strlen (cfg.username) + 1);
			if (! username) {
				fprintf (stderr, "Cannot malloc username!\n");
				return 1;
			}
			strcpy (username, cfg.username);

			password = malloc (strlen (cfg.password) + 1);
			if (! password) {
				fprintf (stderr, "Cannot malloc password!\n");
				return 1;
			}	
			strcpy (password, cfg.password);
		} 
		else {
			username = (text *) my_readline ("Username: ");
			password = (text *) my_readline ("Password: ");
		}

		if (olog (&lda, (ub1 *) hda, username, -1, password, -1,
			(text *) 0, -1, (ub4) OCI_LM_DEF)) {
			if (i >= 2) {
				break;
			}
			if (stdin_is_tty) {
				printf ("Cannot connect to ORACLE as "
					"'%s'...Try Again!\n\n", username);
			}
			else {
				printf ("Cannot connect to ORACLE as '%s'!\n",
					username);
			}
		}
		else {
			if (stdin_is_tty && stdout_is_tty) {
				printf ("You are connected to ORACLE as '%s'\n",
					username);
			}
			free (username);
			free (password);
			return 0;
		}
		free (username);
		free (password);
	}

	printf ("Unable to connect to ORACLE!  Exiting...\n");

	return 1;
}

sword open_cursor (void)
{
	if (oopen (&cda, &lda, (text *) 0, -1, -1, (text *) 0, -1)) {
		fprintf (stderr, "Cannot open cursor!  Exiting...\n");
		ologof (&lda);
		return 1;
        }

	return 0;
}

long cache_data (Cda_Def *cda, sword ncols)
{
	sword col;
	long next_malloc = 0;
	char num_buf[64];
	char ***tmp_row;
	nrows = 0;

	if (! cfg.cache_results) {
		return -1;
	}

	for (col = 0; col < MAX_SELECT_LIST_SIZE; ++col) {
		max_col_len[col] = desc[col].buflen;
	}	

	while (1) {
		if (ofetch (cda)) {
			if (cda->rc == NO_DATA_FOUND) {
				break;
			}
			if (cda->rc != NULL_VALUE_RETURNED) {
				handle_error (cda);
				break;
			}	
	        }

		if (nrows >= next_malloc) {
			next_malloc += 250;
			tmp_row = realloc ((char ***) row, sizeof (void *)
				* next_malloc);
			if (! tmp_row) {
				printf ("cannot realloc at row %ld!\n", nrows);
				free_cache (nrows, ncols);
				return -1;
			}
			row = tmp_row;
		}

//		printf ("about to malloc\n");
		row[nrows] = (char **) malloc (sizeof (void *) * ncols);
		if (! row[nrows]) {
			printf ("cannot malloc row %ld!\n", nrows);
			free_cache (nrows, ncols);
			return -1;
		}

		for (col = 0; col < ncols ; col++) {
			if (def[col].indp < 0) {
				if (4 > max_col_len[col]) { // strlen ("null")
					max_col_len[col] = 4;
				}
				row[nrows][col] = NULL;
				continue;
			}

			switch (desc[col].dbtype) {
			case FLOAT_TYPE:
				sprintf (num_buf, "%0.*f",
					desc[col].scale,
					def[col].flt_buf);
				row[nrows][col] = (char *) malloc
					(strlen (num_buf) + sizeof (char));
				if (! row[nrows][col]) {
					free_cache (nrows, ncols);
					return -1;
				}
				strcpy (row[nrows][col], num_buf);
				break;
			case INT_TYPE:
				sprintf (num_buf, "%d", def[col].int_buf);
				row[nrows][col] = (char *) malloc
					(strlen (num_buf) + sizeof (char));
				if (! row[nrows][col]) {
					free_cache (nrows, ncols);
					return -1;
				}
				strcpy (row[nrows][col], num_buf);
				break;
			default:
				strip_string (def[col].buf);
				row[nrows][col] = (char *) malloc
					(strlen (def[col].buf) + sizeof (char));
				if (! row[nrows][col]) {
					free_cache (nrows, ncols);
					return -1;
				}
				strcpy (row[nrows][col], def[col].buf);
				break;
			}
			if (strlen (row[nrows][col]) > max_col_len[col]) {
				max_col_len[col] = strlen (row[nrows][col]);
			}
		}
		++nrows;
//		printf ("%ld nrows!\n", nrows);
	}

	return nrows;
}

void free_cache (long nrows, sword ncols)
{
	long i, j;

	// free each col
//	printf ("freeing each col\n");
	for (i = 0; i < nrows; ++i) {
		for (j = 0; j < ncols; ++j) {
			if (row[i][j]) {
				free (row[i][j]);
				row[i][j] = NULL;
			}
		}
	}

	// free each row
//	printf ("freeing each row\n");
	for (i = 0; i < nrows; ++i) {
		if (row[i]) {
			free (row[i]);
			row[i] = NULL;
		}
	}

//	printf ("freeing row\n");
	if (row) {
		free (row);
		row = NULL;
	}
} 

sword exec_sql (void)
{
	sword ncols, sql_function;
	struct timeval before, after;
	double secs;

	row = NULL;

	gettimeofday (&before, NULL);

	if (oparse (&cda, (text *) sql_statement, (sb4) -1,
		(sword) PARSE_DEFERRED, (ub4) V7_BEHAVIOR)) {
		handle_error (&cda);
		return 1;
	}

	sql_function = cda.ft;

	if (sql_function == FT_SELECT) {
		if ((ncols = describe_define (&cda)) == -1) {
			return 1;
		}
	}

	if (oexec (&cda)) {
		handle_error (&cda);
		return 1;
	}

	/* if we error on the cache or cfg.cache_results is not		*/
	/* set, run the sql again for the non-cached print <shrug>	*/
	if (sql_function == FT_SELECT) {
		if (cache_data (&cda, ncols) < 1) {
			if (oexec (&cda)) {
				handle_error (&cda);
				return 1;
			}
		}
	}

	gettimeofday (&after, NULL);

	secs = (after.tv_sec + (after.tv_usec / 1000000.00)) -
		(before.tv_sec + (before.tv_usec / 1000000.00));

	if ((sql_function == FT_SELECT) && (row)) {
		switch (display_mode) {
		case MODE_TABLE:
			display_cache_data_table (&cda, nrows, ncols);
			break;
		case MODE_BY_COL:
			display_cache_data_by_col (&cda, nrows, ncols);
			break;
		case MODE_BY_ROW:
			display_cache_data_by_row (&cda, nrows, ncols);
			break;
		default:
			display_cache_data_table (&cda, nrows, ncols);
			break;
		}
		free_cache (nrows, ncols);
	}
	else if (sql_function == FT_SELECT) {
		switch (display_mode) {
		case MODE_TABLE:
			display_data_table (&cda, ncols);
			break;
		case MODE_BY_COL:
			display_data_by_col (&cda, ncols);
			break;
		case MODE_BY_ROW:
			display_data_by_row (&cda, ncols);
			break;
		default:
			display_data_table (&cda, ncols);
			break;
		}
	}

	if (display_timings) {
		if ((sql_function == FT_SELECT) ||
			(sql_function == FT_UPDATE) ||
			(sql_function == FT_DELETE) ||
			(sql_function == FT_INSERT)) {
			printf("%d row%s in set (%0.02f sec%s)\n\n",
				cda.rpc, cda.rpc == 1 ? "" : "s", secs,
				secs == 1.00 ? "" : "s");
		}
		else {
			printf("Statement processed "
				"(%0.02f sec%s)\n\n", secs,
				secs == 1.00 ? " " : "s");
		}
	}

	return 0;
}

void main_loop (void)
{
	sword rc;

	while (1) {
		rc = get_sql_statement ();
		if (rc == 1) {
			if (stdin_is_tty && stdout_is_tty) {
				printf ("Bye!\n");
			}
			close_db ();
			return;
		}
		else if (rc > 1) {
			continue;
		}

		exec_sql ();
	}
}

sword get_sql_statement (void)
{
	text *cbuf;
	text *cp;
	sword lines = 1;
	char prompt[32];

	while (1) {
		if (lines == 1) {
			*sql_statement = '\0';
			sprintf (prompt, "fsql> ");
		}
	        else {
			sprintf (prompt, "   -> ");
		}

		cbuf = (text *) my_readline (prompt);
		if (! cbuf) {
			close_db ();
			exit (0);
		}

		if (*cbuf == '\0') {
			continue;
		}

		if (lines == 1) {
			if ((strncasecmp ((char *) cbuf, "exit", 4) == 0) ||
				(strncasecmp ((char *) cbuf, "quit", 4) == 0) ||
				(strncasecmp ((char *) cbuf, "bye", 3) == 0)) {
				return 1;
			}
			else if (strncasecmp ((char *) cbuf, "env", 3) == 0) {
				display_env ();
				return 2;
			}
			else if (strncasecmp ((char *) cbuf, "help", 4) == 0) {
				display_online_help ();
				return 3;
			}
		}

		if (lines > 1) {
			strcat((char *) sql_statement, " ");
		}
		strcat((char *) sql_statement, (char *) cbuf);

		cp = &sql_statement[strlen((char *) sql_statement) - 1];
        
		while (isspace (*cp)) {
			cp--;
		}
		if (*cp == ';') {
			*cp = '\0';
			break;
		}
		lines++;
	}

	return 0;
}

sword describe_define (Cda_Def *cda)
{
	sword col, deflen, deftyp;
	static ub1 *defptr;

	for (col = 0; col < MAX_SELECT_LIST_SIZE; col++) {
		desc[col].buflen = MAX_ITEM_BUFFER_SIZE;
		if (odescr (cda, col + 1, &desc[col].dbsize,
			&desc[col].dbtype, &desc[col].buf[0],
			&desc[col].buflen, &desc[col].dsize,
			&desc[col].prec, &desc[col].scale,
			&desc[col].nullok)) {
			if (cda->rc == VAR_NOT_IN_LIST) {
				break;
			}
			else {
				handle_error (cda);
				return 1;
			}
		}

		switch (desc[col].dbtype) {
		case NUMBER_TYPE:
			desc[col].dbsize = NUMBER_WIDTH;
			if (desc[col].scale != 0) {
				defptr = (ub1 *) &def[col].flt_buf;
				deflen = (sword) sizeof(float);
				deftyp = FLOAT_TYPE;
				desc[col].dbtype = FLOAT_TYPE;
			}
			else {
				defptr = (ub1 *) &def[col].int_buf;
				deflen = (sword) sizeof(sword);
				deftyp = INT_TYPE;
				desc[col].dbtype = INT_TYPE;
			}
			break;
		default:
			if (desc[col].dbtype == DATE_TYPE) {
				desc[col].dbsize = 32;
			}
			if (desc[col].dbtype == ROWID_TYPE) {
				desc[col].dbsize = 18;
			}
			defptr = def[col].buf;
			deflen = desc[col].dbsize > MAX_ITEM_BUFFER_SIZE ?
				MAX_ITEM_BUFFER_SIZE : desc[col].dbsize + 1;
			deftyp = STRING_TYPE;
			break;
		}

		if (odefin (cda, col + 1, defptr, deflen, deftyp, -1,
			&def[col].indp, (text *) 0, -1, -1,
			&def[col].col_retlen, &def[col].col_retcode)) {
			handle_error (cda);
			return 1;
		}
	}

	return col;
}

void display_col_names_by_row (sword ncols)
{
	sword col;

	for (col = 0; col < ncols ; col++) {
		printf ("%*.*s%c", desc[col].buflen, desc[col].buflen,
			desc[col].buf, (col + 1) == ncols ? '\n' : '\t');
	}
}

void display_col_names_table (sword ncols)
{
	long i, j;

	putchar ('+');
	for (i = 0; i < ncols; ++i) {
		// why can't I figure out how to do this with printf?
		putchar ('-');
		for (j = 0; j < max_col_len[i]; ++j) {
			putchar ('-');
		}
		putchar ('-');
		putchar ('+');
	}
	putchar ('\n');

	for (j = 0; j < ncols; ++j) {
		strip_string (desc[j].buf);
		if ((desc[j].dbtype == FLOAT_TYPE) ||
			(desc[j].dbtype == INT_TYPE)) {
			printf ("| %*s ", max_col_len[j], desc[j].buf);
		}
		else {
			printf ("| %-*s ", max_col_len[j], desc[j].buf);
		}
	}
	puts ("|");
}

void display_data_by_col (Cda_Def *cda, sword ncols)
{
	sword col, t;
	sword max_tablename = 0;
	sword nr = 0;

	for (col = 0; col < ncols; col++) {
		if (desc[col].buflen > max_tablename) {
			max_tablename = desc[col].buflen;
		}
	}

	while (1) {
		if (ofetch (cda)) {
			if (cda->rc == NO_DATA_FOUND) {
				break;
			}
			if (cda->rc != NULL_VALUE_RETURNED) {
				handle_error (cda);
				break;
			}	
	        }

		if (nr > 0) {
			putchar ('\n');
		}
		++nr;

		for (col = 0; col < ncols ; col++) {
			t = 3 + (max_tablename - desc[col].buflen);

			if (def[col].indp < 0) {
				printf ("%*.*s%*cnull\n", 
					desc[col].buflen,
					desc[col].buflen,
					desc[col].buf,	
					t, ' ');
				continue;
			}

			switch (desc[col].dbtype) {
			case FLOAT_TYPE:
				printf ("%*.*s%*c%0.*f\n", 
					desc[col].buflen,
					desc[col].buflen,
					desc[col].buf,	
					t, ' ',
					desc[col].scale,
					def[col].flt_buf);
				break;
			case INT_TYPE:
				printf ("%*.*s%*c%d\n",
					desc[col].buflen,
					desc[col].buflen,
					desc[col].buf,	
					t, ' ',
					def[col].int_buf);
				break;
			default:
				strip_string (def[col].buf);
				printf ("%*.*s%*c%s\n",
					desc[col].buflen,
					desc[col].buflen,
					desc[col].buf,	
					t, ' ',
					def[col].buf);
				break;
			}
		}
		if ((col + 1) < ncols) {
			putchar ('\n');
		}
	}
}

void display_data_by_row (Cda_Def *cda, sword ncols)
{
	sword col;

	if (display_col_names) {
		display_col_names_by_row (ncols);
	}

	while (1) {
		if (ofetch (cda)) {
			if (cda->rc == NO_DATA_FOUND) {
				break;
			}
			if (cda->rc != NULL_VALUE_RETURNED) {
				handle_error (cda);
				break;
			}	
	        }

		for (col = 0; col < ncols ; col++) {
			if (def[col].indp < 0) {
				printf ("\n");
				continue;
			}

			switch (desc[col].dbtype) {
			case FLOAT_TYPE:
				printf ("%0.*f%c", 
					desc[col].scale,
					def[col].flt_buf,
					(col + 1) == ncols ? '\n' : '\t');
				break;
			case INT_TYPE:
				printf ("%d%c",
					def[col].int_buf,
					(col + 1) == ncols ? '\n' : '\t');
				break;
			default:
				strip_string (def[col].buf);
				printf ("%s%c",
					def[col].buf,
					(col + 1) == ncols ? '\n' : '\t');

				break;
			}
		}
	}
}

void display_data_table (Cda_Def *cda, sword ncols)
{
	sword col, n;
	long i, j;

	for (n = 0; n < ncols; ++n) {
		max_col_len[n] = desc[n].dsize;
	}

	if (display_col_names) {
		display_col_names_table (ncols);
	}

	putchar ('+');
	for (i = 0; i < ncols; ++i) {
		// why can't I figure out how to do this with printf?
		putchar ('-');
		for (j = 0; j < max_col_len[i]; ++j) {
			putchar ('-');
		}
		putchar ('-');
		putchar ('+');
	}
	putchar ('\n');

	while (1) {
		if (ofetch (cda)) {
			if (cda->rc == NO_DATA_FOUND) {
				break;
			}
			if (cda->rc != NULL_VALUE_RETURNED) {
				handle_error (cda);
				break;
			}	
	        }

		for (col = 0; col < ncols ; col++) {
			if (def[col].indp < 0) {
				printf ("| %*s ", max_col_len[j], "null");
				continue;
			}
			if ((desc[col].dbtype == FLOAT_TYPE) ||
				(desc[col].dbtype == INT_TYPE)) {
				printf ("| %*s ", max_col_len[j], def[col].buf);
			}
			else {
				printf ("| %-*s ", max_col_len[j],
					def[col].buf);
			}
		}
		puts ("|");
	}

	putchar ('+');
	for (i = 0; i < ncols; ++i) {
		// why can't I figure out how to do this with printf?
		putchar ('-');
		for (j = 0; j < max_col_len[i]; ++j) {
			putchar ('-');
		}
		putchar ('-');
		putchar ('+');
	}
	putchar ('\n');
}

void close_cursor (void)
{
	/*	
	if (orol (&lda)) {
		fprintf (stderr, "Error rolling back transactions!\n");
	}
	*/
	if (oclose (&cda)) {
		fprintf (stderr, "Error closing cursor!\n");
	}
}

void logoff (void)
{
	if (ologof (&lda)) {
		fprintf (stderr, "Error logging off!\n");
	}
}

void close_db (void)
{
	close_cursor ();
	logoff ();
}

void handle_error (Cda_Def *cda)
{
    text msg[512];
    sword n;

    n = oerhms (&lda, cda->rc, msg, (sword) sizeof (msg));
    fprintf (stderr, "%.*s\n", n, msg);
}

void display_cache_data_table (Cda_Def *cda, long nrows, sword ncols)
{
	long i, j;

	if (display_col_names) {
		display_col_names_table (ncols);
	}

	putchar ('+');
	for (i = 0; i < ncols; ++i) {
		// why can't I figure out how to do this with printf?
		putchar ('-');
		for (j = 0; j < max_col_len[i]; ++j) {
			putchar ('-');
		}
		putchar ('-');
		putchar ('+');
	}
	putchar ('\n');

	for (i = 0; i < nrows; ++i) {
		for (j = 0; j < ncols; ++j) {
			if ((desc[j].dbtype == FLOAT_TYPE) ||
				(desc[j].dbtype == INT_TYPE)) {
				if (row[i][j]) {
					printf ("| %*s ", max_col_len[j],
						row[i][j]);
				}
				else {
					printf ("| %*s ", max_col_len[j],
						"null");
				}
			}
			else {
				if (row[i][j]) {
					printf ("| %-*s ", max_col_len[j],
						row[i][j]);
				}
				else {
					printf ("| %-*s ", max_col_len[j],
						"null");
				}
			}
		}
		puts ("|");
	}

	putchar ('+');
	for (i = 0; i < ncols; ++i) {
		// why can't I figure out how to do this with printf?
		putchar ('-');
		for (j = 0; j < max_col_len[i]; ++j) {
			putchar ('-');
		}
		putchar ('-');
		putchar ('+');
	}
	putchar ('\n');
}

void display_cache_data_by_row (Cda_Def *cda, long nrows, sword ncols)
{
	long i, j;

//	printf ("displaying cache data by row\n");

	if (display_col_names) {
		display_col_names_by_row (ncols);
	}

	for (i = 0; i < nrows; ++i) {
		for (j = 0; j < ncols; ++j) {
			if (row[i][j]) {
				printf ("%s%c", row[i][j],
					(j + 1) == ncols ? '\n' : '\t');
			}
			else {
				printf ("null%c", j == (ncols - 1) ?
					'\n' : '\t');
			}
		}
	}	
}

void display_cache_data_by_col (Cda_Def *cda, long nrows, sword ncols)
{
	long i, j;
	sword t;
	sword max_tablename = 0;

	for (i = 0; i < ncols; i++) {
		if (desc[i].buflen > max_tablename) {
			max_tablename = desc[i].buflen;
		}
	}

	for (i = 0; i < nrows; ++i) {
		for (j = 0; j < ncols; ++j) {
			t = 3 + (max_tablename - desc[j].buflen);

			if (row[i][j]) {
				printf ("%*.*s%*c%s\n", 
					desc[j].buflen,
					desc[j].buflen,
					desc[j].buf,	
					t, ' ',
					row[i][j]);
			}
			else {
				printf ("%*.*s%*cnull\n", 
					desc[j].buflen,
					desc[j].buflen,
					desc[j].buf,	
					t, ' ');
			}
		}
		if ((i + 1) < nrows) {
			putchar ('\n');
		}
	}
}
