/*
 * fsql.c - fsf - 2000-02-02
 *  - nicer Oracle interface than sqlplus
 *
 * last modified: 2000-02-06
 *
 * this source is free
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <sys/time.h>
#include <unistd.h>
#include <oratypes.h>
#include <ociapr.h>
#include <ocidem.h>
#include "fsql.h"

extern text sql_statement[MAX_STATEMENT];

char app_name[FILENAME_MAX];
sword stdin_is_tty;
sword stdout_is_tty;
sword display_mode;
sword display_col_names;
sword display_timings;
ConfigStruct cfg;

char *my_readline (char *prompt)
{
	char *line_read = NULL;

	if (stdin_is_tty) {
		line_read = readline (prompt);

		if (line_read && *line_read) {
			add_history (line_read);
		}
	}
	else {
		char buf[256];
		int len = 0;
		char c;
		int got_eol = 1;
	
		while (fread (&c, sizeof (char), 1, stdin)) {
			got_eol = 0;
			if (c == '\n') {
				buf[len] = '\0';
				break;
			}
			else if ((len + 1) < 256) {
				buf[len++] = c;
			}
			else {
				break;
			}
		}
		buf[len] = '\0';

		if (got_eol) {
			close_db ();
			exit (0);
		}

		line_read = (char *) malloc (len + 1);
		if (! line_read) {
			fprintf (stderr, "Cannot malloc in my_readline ()!\n");
			close_db ();
			exit (0);
		}
		strcpy (line_read, buf);
	}

	return (line_read);
}

void check_io (void)
{
	if (isatty (fileno (stdin))) {
		stdin_is_tty = 1;
	}
	else {
		stdin_is_tty = 0;
	}

	if (isatty (fileno (stdout))) {
		stdout_is_tty = 1;
		display_mode = cfg.display_mode_tty;
		display_col_names = cfg.display_col_names_tty;
		display_timings = 1;
	}
	else {
		stdout_is_tty = 0;
		display_mode = cfg.display_mode_notty;
		display_col_names = cfg.display_col_names_notty;
		display_timings = 0;
	}

}

sword exec_arg (char *arg)
{
	sword rc;
	sword i;

	stdin_is_tty = 0;
	display_timings = 0;

	if (logon ()) {
		exit (1);
	}

	if (open_cursor ()) {
		exit (2);
	}

	strcpy (sql_statement, arg);

	for (i = strlen (sql_statement) - 1; i >= 0; --i) {
		if (isspace (sql_statement[i])) {
			sql_statement[i] = '\0';
		}
		else if (sql_statement[i] == ';') {
			sql_statement[i] = '\0';
			break;
		}
		else {
			break;
		}
	}

	rc = exec_sql ();

	close_db ();

	return rc;
}

sword exec_file (char *arg)
{
	sword rc;
	char c;
	sword l = 0;
	FILE *fp;

	display_timings = 0;

	fp = fopen (arg, "r");
	if (! fp) {
		return 3;
	}

	strcpy (sql_statement, arg);

	while (fread (&c, sizeof (c), 1, fp)) {
		if ((c == '\n') && (l < MAX_STATEMENT)) {
			sql_statement[l++] = ' ';
		}
		else if ((c == ';') && (l < MAX_STATEMENT)) {
			sql_statement[l] = '\0';
			// printf ("executing %s\n", sql_statement);
			rc = exec_sql ();
			l = 0;
		}
		else if ((c == ';') && (l >= MAX_STATEMENT)) {
			// reset the sql statement in the event
			// of an overrun
			l = 0;
		}
		else if (l < MAX_STATEMENT) {
			sql_statement[l++] = c;	
		}
	}

	fclose (fp);

	return rc;
}

void display_warranty (void)
{
	printf ("This software comes with ABSOLUTELY NO WARRANTY.\n");
}

void display_version (void)
{
	printf ("fsql v%0.02f\n", APP_VERSION);
}

void display_help (void)
{
	display_version ();
	display_warranty ();
	printf ("\n");
	printf ("Usage: fsql [OPTIONS] [FILES] \n\n");
	printf ("  -c		Override config and display by column\n");
	printf ("  -e		Execute command and quit\n");
	printf ("  -h		Display this help and quit\n");
	printf ("  -n		Display column names when not outputting to a tty\n");
	printf ("  -p		Set the db logon password\n");
	printf ("  -q		Quick mode...do not cache the results\n");
	printf ("  -r		Override config and display by row\n");
	printf ("  -t		Override config and display in table format\n");
	printf ("  -u		Set the db logon username\n");
	printf ("  -v		Display the version and quit\n");
	putchar ('\n');
	printf ("The latest version of fsql can always be found at:\n");
	printf ("http://www.oopdreams.com/frank/fsql\n\n");
}

void get_opts (int argc, char **argv)
{
	int c;
	sword rc = 0;
	sword ec = 0;
	sword i;

	while (1) {
		c = getopt (argc, argv, "ce:hnp:qrtu:v");
		if (c == -1) {
			break;
		}
		switch (c) {
		case '?':
			display_help ();
			exit (1);
		case 'c':
			display_mode = MODE_BY_COL;
			break;
		case 'e':
			if (optarg && (*optarg != '\0')) {
				exit (exec_arg (optarg));
			}
			exit (1);
		case 'h':
			display_help ();
			exit (0);
		case 'n':
			display_col_names = 1;
			break;
		case 'p':
			strncpy (cfg.password, optarg, MAX_PASSWORD - 1);
			break;
		case 'q':
			cfg.cache_results = 0;
			break;
		case 'r':
			display_mode = MODE_BY_ROW;
			break;
		case 't':
			display_mode = MODE_TABLE;
			break;
		case 'u':
			strncpy (cfg.username, optarg, MAX_USERNAME - 1);
			break;
		case 'v':
			display_version ();
			exit (0);
		}
	}

	if (optind < argc) {
		stdin_is_tty = 0;
		display_mode = cfg.display_mode_notty;
		display_col_names = cfg.display_col_names_notty;

		if (logon ()) {
			exit (1);
		}

		if (open_cursor ()) {
			exit (2);
		}

		for (i = optind; i < argc; ++i) {
			// printf ("running %s\n", argv[i]);
			rc = exec_file (argv[i]);
			// just return the highest of the return codes
			if (rc > ec) {
				ec = rc;
			}
		}

		close_db ();
		exit (ec);
	}

}

int main (int argc, char **argv)
{
	strcpy (app_name, argv[0]);

	open_config ();

	check_io ();

	get_opts (argc, argv);

	if (logon ()) {
		exit (1);
	}

	if (open_cursor ()) {
		exit (2);
	}

	main_loop ();

	exit (0);
}
